import os
import re
import json

from .mail import get_mail_program
from .screen import get_screen_program

DEFAULT_CONFIG = dict(
    log_location=os.path.expanduser("~/.log"),
    mail_program="detect",
    screen_program="screen",
)


class Config:
    def __init__(self, config_file):
        config = _load_config(config_file)
        self._config = dict(DEFAULT_CONFIG)
        self._config.update(config)
        self._mail_program = get_mail_program(self._config["mail_program"])
        self._screen_program = get_screen_program(self._config["screen_program"])
        if "email" not in self._config:
            raise RuntimeError(
                "You need to provide an email address, please run `31 config email youraddress@example.com` to set this up"
            )
        self._email = self._config["email"]

    def create_log_file(self, path):
        ll = self._config["log_location"]
        try:
            os.makedirs(ll)
        except FileExistsError:
            pass
        return os.path.join(ll, path)

    def send_mail(self, subject, body):
        return self._mail_program(to=self._email, subject=subject, body=body)

    def launch_screen(self, command, name):
        return self._screen_program(command=command, name=name)


def _load_config(config_file):
    try:
        with open(config_file) as f:
            return json.load(f)
    except FileNotFoundError:
        return {}


def update_config(config_file, new_config):
    config = _load_config(config_file)
    config.update(new_config)
    with open(config_file, "w") as f:
        json.dump(config, f)



import unittest

from .test31 import Test31
from s31.foreach import parse_foreach_statement, parse_foreach_statements


class TestForeachAPI(unittest.TestCase):
    def test_single_statement(self):
        self.assertEqual(
            [[("%x", "1")], [("%x", "2")], [("%x", "3")]],
            parse_foreach_statement(["%x", "1,2,3"]),
        )

    def test_zipped_statement(self):
        self.assertEqual(
            [
                [("%x", "1"), ("%y", "a")],
                [("%x", "2"), ("%y", "b")],
                [("%x", "3"), ("%y", "c")],
            ],
            parse_foreach_statement(["%x", "%y", "1,2,3", "a,b,c"]),
        )

    def test_multiple_statements(self):
        self.assertEqual(
            [
                [("%x", "1"), ("%y", "a")],
                [("%x", "1"), ("%y", "b")],
                [("%x", "2"), ("%y", "a")],
                [("%x", "2"), ("%y", "b")],
            ],
            parse_foreach_statements([["%x", "1,2"], ["%y", "a,b"]]),
        )


class TestForeachCLI(Test31):
    def test_basic_foreach(self):
        self.assertOutput(
            ["31", "c", "--no-email", "-s", "-f", "%x", "1,2,3", "echo %x"],
            ["1", "2", "3", ""],
        )

    def test_csv_foreach(self):
        self.assertOutput(
            ["31", "c", "--no-email", "-s", "-f", "%x", '",",2,3', "echo %x"],
            [",", "2", "3", ""],
        )

    def test_zip_foreach(self):
        self.assertOutput(
            [
                "31",
                "c",
                "--no-email",
                "-s",
                "-f3",
                "%x",
                "%y",
                "%z",
                "1,2,3",
                "a,b,c",
                "x,y,z",
                "echo %x %y %z",
            ],
            ["1 a x", "2 b y", "3 c z", ""],
        )

    def test_prod_foreach(self):
        self.assertOutput(
            [
                "31",
                "c",
                "--no-email",
                "-s",
                "-f",
                "%x",
                "1,2,3",
                "-f",
                "%y",
                "a,b",
                "echo %x %y",
            ],
            ["1 a", "1 b", "2 a", "2 b", "3 a", "3 b", ""],
        )

    def test_screen_foreach(self):
        self.assertOutput(
            [
                "31",
                "c",
                "--no-email",
                "-f",
                "%x",
                "1,2,3",
                "-f",
                "%y",
                "a,b",
                "echo %x %y",
            ],
            [
                "BEGIN SCREEN",
                "NAME = 'echo_1_a'",
                "1 a",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = 'echo_1_b'",
                "1 b",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = 'echo_2_a'",
                "2 a",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = 'echo_2_b'",
                "2 b",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = 'echo_3_a'",
                "3 a",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = 'echo_3_b'",
                "3 b",
                "END SCREEN",
                "",
            ],
        )

    def test_screen_foreach_name(self):
        self.assertOutput(
            [
                "31",
                "c",
                "--no-email",
                "-f",
                "%x",
                "1,2,3",
                "-f",
                "%y",
                "a,b",
                "-n %x_%y",
                "echo %x %y",
            ],
            [
                "BEGIN SCREEN",
                "NAME = '1_a'",
                "1 a",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = '1_b'",
                "1 b",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = '2_a'",
                "2 a",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = '2_b'",
                "2 b",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = '3_a'",
                "3 a",
                "END SCREEN",
                "BEGIN SCREEN",
                "NAME = '3_b'",
                "3 b",
                "END SCREEN",
                "",
            ],
        )



import sys
import subprocess
import re


def get_screen_program(name):
    if name == "screen":
        return _screen
    if name == "test":
        return _test
    else:
        raise RuntimeError("Only screen is currently supported as a screen_program")


def _screen(command, name):
    name = re.sub(r"[^A-Za-z0-9-., ]", "?", name)
    screen_command = ["screen", "-S", name, "-d", "-m", "--", *command]
    subprocess.check_call(screen_command)


def _test(command, name):
    print("BEGIN SCREEN")
    print("NAME =", repr(name))
    sys.stdout.flush()
    subprocess.run(command)
    sys.stdout.flush()
    print("END SCREEN")
    sys.stdout.flush()



import attr
import sys
import subprocess


from .utils import format_assignments


@attr.s
class Command:
    cmd_line = attr.ib()
    location = attr.ib()

    def run_teed(self, logfile):
        with open(logfile, "wb") as f:

            def write(x):
                f.write(x)
                sys.stdout.buffer.write(x)
                f.flush()
                sys.stdout.flush()

            p = subprocess.Popen(
                self.cmd_line,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                **self.kwargs
            )
            while p.poll() is None:
                line = p.stdout.readline()
                if line:
                    write(line)
            while True:
                line = p.stdout.readline()
                if line:
                    write(line)
                else:
                    break
            return p.returncode

    def replace(self, assignments):
        return Command(
            cmd_line=format_assignments(self.cmd_line, assignments),
            location=format_assignments(self.location, assignments),
        )

    @property
    def kwargs(self):
        kwargs = dict(shell=1, bufsize=0)
        if self.location is not None:
            kwargs["cwd"] = self.location
        return kwargs

    def run(self):
        subprocess.run(self.cmd_line, **self.kwargs)

    def dry_run(self):
        print(self.cmd_line)



import subprocess
import re

import shelve
from filelock import FileLock


def output_matches(command, regex):
    output = subprocess.run(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ).stdout
    return re.match(regex, output.decode("utf-8"))


def format_assignments(value, assignments):
    if value is None:
        return value
    for var, val in assignments:
        value = value.replace(var, val)
    return value


def sanitize(name):
    return re.sub("[^A-Za-z0-9_.]+", "_", name).strip("_")


def set_key(file, key):
    with FileLock(file + ".lock"):
        with shelve.open(file + ".shelve") as s:
            s[key] = True


def get_keys(file):
    with FileLock(file + ".lock"):
        with shelve.open(file + ".shelve") as s:
            return {k for k in s if s[k]}



import tempfile
import uuid
import time

from .utils import get_keys


def dispatch_workers(num_workers, launch_worker, arguments):
    used_workers = [None] * num_workers
    arguments = arguments[::-1]  # so we can pop from the back
    completed = 0
    file = tempfile.mktemp()
    while True:
        # check if any workers have completed
        dones = get_keys(file)
        for i, x in enumerate(used_workers):
            if x in dones:
                completed += 1
                used_workers[i] = None

        free = [i for i, w in enumerate(used_workers) if w is None]

        if len(free) == num_workers and not arguments:
            print("All tasks completed")
            return

        print(
            "Completed: {}, In Progress: {}, Queued: {}".format(
                completed, num_workers - len(free), len(arguments)
            )
        )

        if not free or not arguments:
            # Wait for some process to become free, or just wait for everything to end
            time.sleep(1)
            continue

        worker_idx = free[0]
        used_workers[worker_idx] = str(uuid.uuid4())
        launch_worker(worker_idx, arguments.pop(), file, used_workers[worker_idx])



from .test31 import Test31
import os


class TestOptions(Test31):
    def test_sync(self):
        self.assertOutput(
            ["31", "c", "--sync", "echo test"],
            [
                "test",
                "SENDING EMAIL",
                "TO = 'test@example.com'",
                "SUBJECT = 'Process succeeded in 0 seconds: echo test'",
                "BODY = 'test\\n'",
                "",
            ],
        )

    def test_name(self):
        self.assertOutput(
            ["31", "c", "-n", "testing command", "echo test"],
            [
                "BEGIN SCREEN",
                "NAME = 'testing_command'",
                "test",
                "SENDING EMAIL",
                "TO = 'test@example.com'",
                "SUBJECT = 'Process succeeded in 0 seconds: echo test'",
                "BODY = 'test\\n'",
                "END SCREEN",
                "",
            ],
        )

    def test_no_emails(self):
        self.assertOutput(
            ["31", "c", "--no-email", "echo test"],
            [
                "BEGIN SCREEN",
                "NAME = 'echo_test'",
                "test",
                "END SCREEN",
                "",
            ],
        )

    def test_pwd(self):
        self.assertOutput(
            [
                "31",
                "c",
                "-l",
                "..",
                "-n",
                "testing command",
                "python -c 'import os; print(os.getcwd())'",
            ],
            [
                "BEGIN SCREEN",
                "NAME = 'testing_command'",
                os.path.abspath(".."),
                "SENDING EMAIL",
                "TO = 'test@example.com'",
                "SUBJECT = \"Process succeeded in 0 seconds: python -c 'import os; print(os.getcwd())'\"",
                "BODY = '{}\\n'".format(os.path.abspath("..")),
                "END SCREEN",
                "",
            ],
        )



from datetime import datetime
from display_timedelta import display_timedelta

FORMAT = "%Y.%m.%d.%H.%M.%S.%f"


def notify(config, command):
    start = datetime.now()
    path = datetime.strftime(start, FORMAT)
    logfile = config.create_log_file(path)

    exitcode = command.run_teed(logfile)

    end = datetime.now()

    with open(logfile, "rb") as f:
        log_contents = f.read()

    delta = end - start
    subject = "{} in {}: {}".format(
        "Process succeeded"
        if exitcode == 0
        else "Process failed with code {}".format(exitcode),
        display_timedelta(delta).replace("right now", "0 seconds"),
        command.cmd_line,
    )
    config.send_mail(subject, log_contents)



from .test31 import Test31


class BasicTest(Test31):
    def test_success(self):
        self.assertOutput(
            ["31", "c", "echo test"],
            [
                "BEGIN SCREEN",
                "NAME = 'echo_test'",
                "test",
                "SENDING EMAIL",
                "TO = 'test@example.com'",
                "SUBJECT = 'Process succeeded in 0 seconds: echo test'",
                "BODY = 'test\\n'",
                "END SCREEN",
                "",
            ],
        )

    def test_failure(self):
        self.assertOutput(
            ["31", "c", "exit 17"],
            [
                "BEGIN SCREEN",
                "NAME = 'exit_17'",
                "SENDING EMAIL",
                "TO = 'test@example.com'",
                "SUBJECT = 'Process failed with code 17 in 0 seconds: exit 17'",
                "BODY = ''",
                "END SCREEN",
                "",
            ],
        )

    def test_multiline_output(self):
        self.assertOutput(
            ["31", "c", "echo 2; echo 3"],
            [
                "BEGIN SCREEN",
                "NAME = 'echo_2_echo_3'",
                "2",
                "3",
                "SENDING EMAIL",
                "TO = 'test@example.com'",
                "SUBJECT = 'Process succeeded in 0 seconds: echo 2; echo 3'",
                "BODY = '2\\n3\\n'",
                "END SCREEN",
                "",
            ],
        )

    def test_stdout_stderr_output(self):
        self.assertOutput(
            ["31", "c", "echo 2; python3 -c 'import sys; print(3, file=sys.stderr)'"],
            [
                "BEGIN SCREEN",
                "NAME = 'echo_2_python3_c_import_sys_print_3_file_sys.stderr'",
                "2",
                "3",
                "SENDING EMAIL",
                "TO = 'test@example.com'",
                "SUBJECT = \"Process succeeded in 0 seconds: echo 2; python3 -c 'import sys; print(3, file=sys.stderr)'\"",
                "BODY = '2\\n3\\n'",
                "END SCREEN",
                "",
            ],
        )



import argparse
import sys
import os
import json

from .config import Config, update_config
from .notify import notify
from .command import Command
from .foreach import MAX_FOREACHES, parse_foreach_args, parse_values
from .utils import format_assignments, sanitize, set_key
from .workers import dispatch_workers


def main():
    def config_argument(p):
        p.add_argument(
            "--config-file",
            default=os.path.expanduser("~/.31rc"),
            help="The location of the configuration file",
        )

    parser = argparse.ArgumentParser("31")
    subparsers = parser.add_subparsers(dest="cmd")
    subparsers.required = True

    command_parser = subparsers.add_parser(
        "command", help="Run a command", aliases=["c"]
    )
    config_argument(command_parser)
    command_parser.add_argument(
        "-s",
        "--sync",
        action="store_true",
        help="Run the command synchronously, that is, not in a screen session",
    )
    command_parser.add_argument(
        "-n", "--screen-name", help="The name of the screen session to create"
    )
    command_parser.add_argument(
        "-l", "--location", help="The location to run the script"
    )
    command_parser.add_argument(
        "--no-email",
        help="Do not send an email when the command is done running",
        action="store_true",
    )
    command_parser.add_argument(
        "-d",
        "--dry-run",
        help="Print out the commands to be run rather than running them",
        action="store_true",
    )
    command_parser.add_argument(
        "-f",
        "--foreach",
        metavar=("%var", "vals"),
        nargs=2,
        action="append",
        help="Replaces each occurence of the variable with the corresponding value. "
        "Variables can be any sequence of characters. "
        "After the variables, values can be provided, each list of values should be a single argument in CSV format. "
        "See the documentation for details and examples.",
    )
    command_parser.add_argument(
        "-fw",
        "--foreach-worker",
        nargs=2,
        metavar=("%var", "vals"),
        action="append",
        help="Similar to -f but associates each substitution with a particular variable. "
        "Notably, this does not lead to a combinatoric explosion if multiple are used, they are "
        "implicitly zipped together by the worker index",
    )
    for k in range(2, 1 + MAX_FOREACHES):
        meta = tuple("%var{}".format(i) for i in range(1, k + 1))
        meta += tuple("vals{}".format(i) for i in range(1, k + 1))
        command_parser.add_argument(
            "-f" + str(k),
            "--foreach-" + str(k),
            metavar=meta,
            nargs=k * 2,
            action="append",
            help="See -f for details, -f2 through -f{0} allow you to zip the values for 2-{0} variables together.".format(
                MAX_FOREACHES
            )
            if k == 2
            else argparse.SUPPRESS,
        )
    # internal use only, specifies which foreach args to use, in json format [(name, value)]
    command_parser.add_argument(
        "--foreach-specified-args", type=json.loads, help=argparse.SUPPRESS
    )
    command_parser.add_argument(
        "-w",
        "--max-workers",
        type=int,
        help="Limit the number of threads that are to be launched at any point. "
        "This forces the creation of a monitoring thread, for which --sync is applied to",
    )
    command_parser.add_argument(
        "-wn",
        "--worker-monitor-name",
        default="worker-monitor",
        help="Names the screen for the worker thread. By default is 'worker-monitor'",
    )
    # internal use only, specifies that when the process is done, it should set the given key of the given file.
    command_parser.add_argument(
        "--when-done-set", nargs=2, metavar=("file", "key"), help=argparse.SUPPRESS
    )
    command_parser.add_argument("command", help="Command to run")
    command_parser.set_defaults(action=command_action)

    config_parser = subparsers.add_parser("config", help="Modify configuration")
    config_argument(config_parser)
    config_parser.add_argument("key", help="The configuration key to modify")
    config_parser.add_argument("value", help="The value to assign the given key to")
    config_parser.set_defaults(action=config_action)
    args = parser.parse_args()

    try:
        args.action(args)
    except RuntimeError as e:
        print(e, file=sys.stderr)


def command_action(args):
    try:
        return do_command_action(args)
    finally:
        if args.when_done_set is not None:
            set_key(*args.when_done_set)


def do_command_action(args):
    config = Config(args.config_file)
    assignments = (
        [args.foreach_specified_args]
        if args.foreach_specified_args is not None
        else parse_foreach_args(args)
    )
    worker_assignments = []
    # Validation
    if args.foreach_worker is not None:
        for variable, vals in args.foreach_worker:
            if args.max_workers is None:
                raise RuntimeError("Cannot provide -fw without -w")
            vals = parse_values(vals)
            if len(vals) != args.max_workers:
                raise RuntimeError(
                    "Mismatch between number of workers and number of provided values"
                )
            worker_assignments.append((variable, vals))

    commands = []
    for assignment in assignments:
        screen_name = sanitize(
            format_assignments(args.screen_name or args.command, assignment)
        )
        cmd = Command(cmd_line=args.command, location=args.location)
        cmd_to_use = cmd.replace(assignment)
        commands.append((screen_name, cmd_to_use, assignment))

    if args.dry_run:
        for screen_name, cmd_to_use, _ in commands:
            if not args.sync:
                print("# on screen {}".format(screen_name))
            cmd_to_use.dry_run()
        return

    if not args.sync:
        if args.max_workers is not None:
            config.launch_screen(sys.argv + ["--sync"], args.worker_monitor_name)
            return
        for screen_name, _, assignment in commands:
            config.launch_screen(
                sys.argv
                + ["--sync", "--foreach-specified-args", json.dumps(assignment)],
                screen_name,
            )
        return

    if args.max_workers is not None and args.foreach_specified_args is None:
        # if foreach-specified-args is set, this is a dispatch thread
        def launch_worker(worker_idx, assignment, output_file, token):
            assignment = assignment + [
                (var, vals[worker_idx]) for var, vals in worker_assignments
            ]
            screen_name = sanitize(
                format_assignments(args.screen_name or args.command, assignment)
            )
            print("Launching {}".format(screen_name))
            # don't need the --sync since at this point it is guaranteed
            config.launch_screen(
                sys.argv
                + [
                    "--foreach-specified-args",
                    json.dumps(assignment),
                    "--when-done-set",
                    output_file,
                    token,
                ],
                screen_name,
            )

        dispatch_workers(args.max_workers, launch_worker, assignments)
        return

    for _, cmd_to_use, _ in commands:
        if args.no_email:
            cmd_to_use.run()
        else:
            notify(config, cmd_to_use)


def config_action(args):
    update_config(args.config_file, {args.key: args.value})



import os
import subprocess
import tempfile

import unittest

RC_TEST = os.path.join(os.path.dirname(__file__), "testrc.json")


class Test31(unittest.TestCase):
    def assertOutput(self, command, output, check=True, **kwargs):
        path = tempfile.mktemp()
        with open(path, "wb") as f:
            try:
                subprocess.run(
                    [*command, "--config-file", RC_TEST],
                    stdout=f,
                    stderr=subprocess.STDOUT,
                    check=check,
                    **kwargs
                )
            except subprocess.TimeoutExpired:
                pass
        with open(path) as f:
            self.assertEqual(f.read().split("\n"), output)



import itertools
import csv

MAX_FOREACHES = 9


def parse_foreach_args(args):
    foreach = []
    if args.foreach is not None:
        foreach += args.foreach
    for k in range(2, 1 + MAX_FOREACHES):
        foreach_k = getattr(args, "foreach_{}".format(k))
        if foreach_k is not None:
            foreach += foreach_k
    return parse_foreach_statements(foreach)


def parse_foreach_statements(foreach):
    all_assigns = [
        list(itertools.chain(*x))
        for x in itertools.product(
            *[parse_foreach_statement(statement) for statement in foreach]
        )
    ]
    return all_assigns


def parse_foreach_statement(statement):
    assert statement, "must provide at least one element"
    assert len(statement) % 2 == 0, "must provide an even number of elements"
    k = len(statement) // 2
    variables = statement[:k]
    values = statement[k:]
    values = [parse_values(v) for v in values]
    for v in values:
        if len(v) != len(values[0]):
            raise RuntimeError(
                "Values must be the same length but have lengths of {} ({!r}) vs {} ({!r})".format(
                    len(values[0]), values[0], len(v), v
                )
            )
    return [
        [(variable, v) for variable, v in zip(variables, vals)]
        for vals in list(zip(*values))
    ]


def parse_values(values):
    return list(csv.reader([values]))[0]



import subprocess
import sys

from .utils import output_matches


def get_mail_program(program_name):
    if program_name == "detect":
        if output_matches("mail -V", "^mail \(GNU Mailutils\).*"):
            program_name = "gnu_mail"
        elif output_matches("mutt -h", "Mutt .*"):
            program_name = "mutt"
        else:
            raise RuntimeError(
                "Could not detect a mail program. Please see the documentation for a list of supported programs."
            )
    return dict(gnu_mail=_gnu_mail, mutt=_mutt, test=_test)[program_name]


def _gnu_mail(to, subject, body):
    subprocess.run(["mail", "-s", subject, to], input=body)


def _mutt(to, subject, body):
    subprocess.run(["mutt", "-s", subject, to], input=body)


def _test(to, subject, body):
    print("SENDING EMAIL")
    print("TO =", repr(to))
    print("SUBJECT =", repr(subject))
    print("BODY =", repr(body.decode("utf-8")))
    sys.stdout.flush()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="31",
    version="1.4",
    author="Kavi Gupta",
    author_email="31@kavigupta.org",
    description="Runs code in a specified environment in the background and notifies you when it is done.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kavigupta/31",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": ["31=s31.main:main"],
    },
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
    install_requires=["attrs>=20.1.0", "display-timedelta==1.1", "filelock==3.0.12"],
)



from .test31 import Test31


class BasicTest(Test31):
    def test_success(self):
        self.assertOutput(
            [
                "31",
                "c",
                "-s",
                "--no-email",
                'python -u -c "import time, itertools; [(print(k), time.sleep(2)) for k in itertools.count()]"',
            ],
            [
                "0",
                "1",
                "2",
                "",
            ],
            check=0,
            timeout=5,
        )
